package edu.illinois.cs.cs124.ay2024.mp.test

import android.content.Intent
import android.widget.ToggleButton
import androidx.test.core.app.ApplicationProvider
import androidx.test.espresso.Espresso.onView
import androidx.test.espresso.assertion.ViewAssertions.matches
import androidx.test.espresso.matcher.ViewMatchers.hasDescendant
import androidx.test.espresso.matcher.ViewMatchers.isAssignableFrom
import androidx.test.espresso.matcher.ViewMatchers.isDisplayed
import androidx.test.espresso.matcher.ViewMatchers.withId
import androidx.test.espresso.matcher.ViewMatchers.withText
import androidx.test.ext.junit.runners.AndroidJUnit4
import com.fasterxml.jackson.databind.JsonNode
import com.fasterxml.jackson.databind.node.ObjectNode
import com.google.common.truth.Truth.assertWithMessage
import edu.illinois.cs.cs124.ay2024.mp.R
import edu.illinois.cs.cs124.ay2024.mp.activities.RSOActivity
import edu.illinois.cs.cs124.ay2024.mp.models.Favorite
import edu.illinois.cs.cs124.ay2024.mp.models.RSO
import edu.illinois.cs.cs124.ay2024.mp.models.Summary
import edu.illinois.cs.cs124.ay2024.mp.network.Client
import edu.illinois.cs.cs124.ay2024.mp.network.resetServer
import edu.illinois.cs.cs124.ay2024.mp.network.startServer
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.GET_METHOD_EXTRA_TIME
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.JSONReadCountSecurityManager
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.RSODATA
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.SUMMARIES
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.SUMMARY_COUNT
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.configureLogging
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.countRecyclerView
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.doCheckServerDesign
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.getShuffledSummaries
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.isChecked
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.objectMapper
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.pause
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.setChecked
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.startActivity
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.startMainActivity
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testClient
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testServerGet
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testServerGetTimed
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testServerPost
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.trimmedMean
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.withRecyclerView
import edu.illinois.cs.cs125.gradlegrader.annotations.Graded
import org.hamcrest.Matchers.containsString
import org.junit.After
import org.junit.AfterClass
import org.junit.Before
import org.junit.BeforeClass
import org.junit.FixMethodOrder
import org.junit.Test
import org.junit.runner.RunWith
import org.junit.runners.MethodSorters
import org.robolectric.annotation.LooperMode
import org.robolectric.annotation.experimental.LazyApplication
import java.net.HttpURLConnection
import java.util.Random

/*
 * This is the MP3 test suite.
 * The code below is used to evaluate your app during testing, local grading, and official grading.
 * You may not understand all of the code below, but you'll need to have some understanding of how
 * it works so that you can determine what is wrong with your app and what you need to fix.
 *
 * ALL CHANGES TO THIS FILE WILL BE OVERWRITTEN DURING OFFICIAL GRADING.
 * You can and should modify the code below if it is useful during your own local testing,
 * but any changes you make will be discarded during official grading.
 * The local grader will not run if the test suites have been modified, so you'll need to undo any
 * local changes before you run the grader.
 *
 * Note that this means that you should not fix problems with the app by modifying the test suites.
 * The test suites are always considered to be correct.
 *
 * Our test suites are broken into two parts.
 * The unit tests are tests that we can perform without running your app.
 * They test things like whether a method works properly or the behavior of your API server.
 * Unit tests are usually fairly fast.
 *
 * The integration tests are tests that require simulating your app.
 * This allows us to test things like your API client, and higher-level aspects of your app's
 * behavior, such as whether it displays the right thing on the display.
 * Because integration tests require simulating your app, they run more slowly.
 *
 * The MP3 test suite includes no ungraded tests.
 * These tests are fairly idiomatic, in that they resemble tests you might write for an actual
 * Android programming project.
 * Not entirely though.
 */

@RunWith(AndroidJUnit4::class)
@LooperMode(LooperMode.Mode.PAUSED)
@FixMethodOrder(MethodSorters.NAME_ASCENDING)
class MP3Test {

    // Test the GET and POST /favorite server routes
    @SuppressWarnings("SpellCheckingInspection")
    @Test(timeout = 60000L)
    @Graded(points = 20, friendlyName = "Server POST /favorite (Unit)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test0_ServerGETAndPOSTFavorite() {
        // Get a shuffled list of summaries
        val random = Random(12431)
        val trimmedSummaries = getShuffledSummaries(12431, 128)

        // Perform initial GET /favorite requests
        for (summary in trimmedSummaries) {
            // Perform initial GET
            val favorite = "/favorite/${summary.id}".testServerGet<Favorite>(Favorite::class.java)!!
            // No RSOs should be a favorite at this point
            assertWithMessage("Incorrect favorite value for RSO")
                .that(favorite.favorite)
                .isEqualTo(false)
        }

        // Perform POST /favorite requests to change favorite status

        // Map to store the favorite status set during the next round
        val favorites = mutableMapOf<String, Boolean>()

        // Reshuffle the list of summaries
        for (summary in trimmedSummaries.shuffled(random)) {
            // POST to change favorite value
            val isFavorite = random.nextBoolean()

            // Construct POST favorite body
            val newFavorite = objectMapper.createObjectNode().apply {
                set<JsonNode>("id", objectMapper.convertValue(summary.id, JsonNode::class.java))
                set<JsonNode>("favorite", objectMapper.convertValue(isFavorite, JsonNode::class.java))
            }

            // POST to change the favorite status accordingly
            val favorite = "/favorite".testServerPost<Favorite>(newFavorite, Favorite::class.java)!!
            // Ensure the result is what we expect
            assertWithMessage("Incorrect value from favorite POST")
                .that(favorite.favorite)
                .isEqualTo(isFavorite)

            // Save favorite value for next stage
            favorites[summary.id] = isFavorite
        }

        // Save response times for comparison
        val baseResponseTimes = mutableListOf<Long>()
        val getResponseTimes = mutableListOf<Long>()

        // Second route of GET /favorite requests to ensure favorite status is saved
        for (summary in trimmedSummaries.shuffled(random)) {
            // Time the index route for comparison
            val baseResult = "/".testServerGetTimed<String>(String::class.java)
            baseResponseTimes.add(baseResult.responseTime.toNanos())

            // Retrieve saved favorite
            val savedFavorite = favorites[summary.id]!!

            // Final GET
            val favoriteResult = "/favorite/${summary.id}".testServerGetTimed<Favorite>(Favorite::class.java)
            // Ensure the result is what we expect
            assertWithMessage("Incorrect favorite value for RSO")
                .that(favoriteResult.response!!.favorite)
                .isEqualTo(savedFavorite)
            getResponseTimes.add(favoriteResult.responseTime.toNanos())
        }

        // Check for slow server GETs potentially caused by unnecessary parsing
        val averageBase = trimmedMean(baseResponseTimes, 0.1)
        val averageResponse = trimmedMean(getResponseTimes, 0.1)
        assertWithMessage("Server GET /favorite/ is too slow")
            .that(averageResponse)
            .isLessThan(averageBase * GET_METHOD_EXTRA_TIME)

        // Test bad requests

        // Non-existent RSO GET
        "/favorite/XgRG0-qzi7cNaYPtF8lI1m62tEO/".testServerGet(HttpURLConnection.HTTP_NOT_FOUND)

        // Bad URL GET
        "/favorites/8OozGTNQpaETr0sAB4UVt_ee2qk".testServerGet(HttpURLConnection.HTTP_NOT_FOUND)

        // Non-existent RSO POST
        @Suppress("SpellCheckingInspection")
        val nonexistentRSO: ObjectNode = objectMapper.createObjectNode().apply {
            set<JsonNode>("id", objectMapper.convertValue("ejEQxt9QtbWpjqBIvMt4XaxDDdW", JsonNode::class.java))
            set<JsonNode>("favorite", objectMapper.convertValue(false, JsonNode::class.java))
        }
        "/favorite".testServerPost(nonexistentRSO, HttpURLConnection.HTTP_NOT_FOUND)

        // Bad URL POST
        "/favorites/".testServerPost(nonexistentRSO, HttpURLConnection.HTTP_NOT_FOUND)

        // Bad body POST
        val badBody: ObjectNode = objectMapper.createObjectNode().apply {
            set<JsonNode>("id", objectMapper.convertValue("8OozGTNQpaETr0sAB4UVt_ee2qk", JsonNode::class.java))
            set<JsonNode>("isFavorite", objectMapper.convertValue(false, JsonNode::class.java))
        }
        "/favorite".testServerPost(badBody, HttpURLConnection.HTTP_BAD_REQUEST)
    }

    // Test the client getFavorite and setFavorite methods
    @Test(timeout = 30000L)
    @Graded(points = 20, friendlyName = "Client getFavorite and setFavorite (Integration)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test1_ClientGetAndSetFavorite() {
        // Get a shuffled list of summaries
        val random = Random(12433)
        val trimmedSummaries = getShuffledSummaries(12433, 128)
        // Map to store the favorite status set by Client.setFavorite
        val favorites = mutableMapOf<String, Boolean>()

        // Go through all RSOs twice
        repeat(2) {
            for (summary in trimmedSummaries) {
                // Randomly either getFavorite or setFavorite
                var currentFavorite: Boolean
                if (random.nextBoolean()) {
                    currentFavorite =
                        testClient { callback -> Client.getFavorite(summary.id, callback) }
                } else {
                    val isFavorite = random.nextBoolean()
                    favorites[summary.id] = isFavorite
                    currentFavorite =
                        testClient { callback -> Client.setFavorite(summary.id, isFavorite, callback) }
                }

                // Check against the expected value
                val expectedFavorite = favorites.getOrDefault(summary.id, false)
                assertWithMessage("Incorrect favorite value")
                    .that(currentFavorite)
                    .isEqualTo(expectedFavorite)
            }
        }

        // Test bad Client.getFavorite call
        try {
            testClient { callback -> Client.getFavorite("Lr363RRTGsnNHO3EO37VkKOk65a", callback) }
            assertWithMessage("Client GET /favorite for non-existent RSO should throw").fail()
        } catch (ignored: Exception) {
        }
    }

    // Helper method for the favorite button UI test
    private fun favoriteButtonHelper(summary: Summary?, currentFavorite: Boolean, nextFavorite: Boolean) {
        // Prepare the Intent to start the RSOActivity
        val intent = Intent(ApplicationProvider.getApplicationContext(), RSOActivity::class.java).apply {
            putExtra("id", summary!!.id)
        }

        // Start the RSOActivity
        startActivity<RSOActivity>(intent) {
            pause()
            // Check that the initial favorite status is correct, change it,
            // and then verify the change
            onView(isAssignableFrom(ToggleButton::class.java))
                .check(isChecked(currentFavorite))
                .perform(setChecked(nextFavorite))
                .check(isChecked(nextFavorite))
        }
    }

    // Test the favorite button
    @Test(timeout = 30000L)
    @Graded(points = 15, friendlyName = "Favorite Button (Integration)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test2_FavoriteButton() {
        // Get a shuffled small list of summaries
        val trimmedSummaries = getShuffledSummaries(12434, 2)
        val random = Random(12434)
        // Map to store our expected favorite results
        val favorites = mutableMapOf<String, Boolean>()

        // Go through each summary four times, either setting or clearing its favorite status
        repeat(4) {
            for (summary in trimmedSummaries.shuffled(random)) {
                val currentFavorite = favorites.getOrDefault(summary.id, false)
                val nextFavorite = random.nextBoolean()
                favorites[summary.id] = nextFavorite
                favoriteButtonHelper(summary, currentFavorite, nextFavorite)
            }
        }
    }

    // Helper method for the favorite sort UI test
    private fun favoriteSortHelper(favoriteSummaries: List<Summary>) {
        // Start the main activity
        startMainActivity {
            // Check that the right number of RSOs are displayed initially
            onView(withId(R.id.recycler_view)).check(countRecyclerView(SUMMARY_COUNT))

            // Check that the expected RSOs are at the top of the list
            for (position in favoriteSummaries.indices) {
                val title = favoriteSummaries[position].title
                onView(withRecyclerView(R.id.recycler_view).atPosition(position))
                    .check(matches(hasDescendant(withText(title))))
            }
        }
    }

    // Test the favorite sort
    @Test(timeout = 30000L)
    @Graded(points = 15, friendlyName = "Favorite Sort (Integration)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test3_FavoriteSort() {
        // Get a shuffled small list of summaries
        val trimmedSummaries = getShuffledSummaries(12435, 4)
        // Track which summaries are current favorites
        val favoriteSummaries = mutableListOf<Summary>()

        // Go through each summary, marking it is a favorite, and then making sure the main activity
        // UI is updated properly
        for (summary in trimmedSummaries) {
            favoriteButtonHelper(summary, currentFavorite = false, nextFavorite = true)
            favoriteSummaries.add(summary)
            favoriteSortHelper(favoriteSummaries.sortedBy { it.title })
        }
    }

    private fun relatedRSOsHelper(id: String, expectedSize: Int, expectedIDs: List<String>) {
        val rso = "/rso/$id".testServerGet<RSO>(RSO::class.java)!!
        assertWithMessage("Wrong number of related RSOs")
            .that(rso.relatedRSOs)
            .hasSize(expectedSize)
        for (i in 0 until expectedSize) {
            assertWithMessage("Related results mismatch at index $i")
                .that(rso.relatedRSOs[i].id)
                .startsWith(expectedIDs[i])
        }
    }

    // Test that the RSO related RSOs field is set properly
    @Test(timeout = 30000L)
    @Graded(points = 5, friendlyName = "Related RSOs (Unit)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    @Suppress("LongMethod", "SpellCheckingInspection")
    fun test4_testRelatedRSOs() {
        // These test cases were generated from the reference solution
        relatedRSOsHelper(
            "32-4gqW-bwXw-kWCkEOUcRHPUs4",
            8,
            listOf("in6m", "rKUE", "iVc3", "5WQ1", "C75l", "_5cc", "hvLA", "1miu"),
        )
        relatedRSOsHelper(
            "rSTwDZv3pzBZJKaPBYobW4NS9Eg",
            8,
            listOf("eQd5", "ciZG", "1miu", "5gES", "Z9L6", "H0DT", "HmT2", "B0D8"),
        )
        relatedRSOsHelper(
            "HJAfkmSOYPES7L67HtpjUxxCyHw",
            8,
            listOf("wlsU", "Iwu0", "rVWl", "Ply2", "ZEVD", "yZ8P", "Lu9I", "Mtjc"),
        )
        relatedRSOsHelper(
            "W1xasD7GZprE8HXhJHiIo9F5c6I",
            8,
            listOf("9oqb", "KPpO", "0Ehk", "061b", "SX3L", "hdeL", "90WJ", "g27m"),
        )
        relatedRSOsHelper("5HyU4jaDURz5tJkfMk_CTKng6l4", 0, emptyList())
        relatedRSOsHelper(
            "urdk_DYN0grLi-aPBpqfZcf9c7M",
            8,
            listOf("nSms", "h836", "9oqb", "8nVP", "AhFn", "Bo1L", "FV6P", "Wz-m"),
        )
        relatedRSOsHelper("Ycf2sRS-kbyupsVpszeay7w8DdQ", 0, emptyList())
        relatedRSOsHelper(
            "gxFlRlrhUjaYy73v-Qpp4N4on58",
            8,
            listOf("qWc7", "nSms", "SvzT", "_fSS", "h836", "oRxw", "9oqb", "l5U_"),
        )
        relatedRSOsHelper(
            "L2QwNEcIdpkEQAlt1BMpbVgUS1M",
            8,
            listOf("h836", "7zCc", "Z0o5", "9oqb", "5WQ1", "lWVa", "oHiC", "ciZG"),
        )
        relatedRSOsHelper("CRyYD8vzu_MomRwOy28FpJF31ms", 0, emptyList())
        relatedRSOsHelper(
            "AzBhqKVMK-57c02gXru8S4EtBks",
            8,
            listOf("eQd5", "XuEp", "EBpj", "1b-2", "zwzD", "NG7P", "yMXt", "GwoD"),
        )
        relatedRSOsHelper(
            "-cY_Ba4lVngvoj1YtKDGD3HL8Sw",
            8,
            listOf("q-_r", "V4O3", "uUgY", "NAzx", "rVWl", "Ply2", "ZEVD", "6lVH"),
        )
        relatedRSOsHelper("IzASOTLG6b9TlkFvO3kgJw5vxFs", 0, emptyList())
        relatedRSOsHelper("aJIEVLD9SAq8-CuG7LqTuT7NNrQ", 0, emptyList())
        relatedRSOsHelper(
            "ck_fqfjhHdRSNFzO3k2EgOYIXZg",
            8,
            listOf("QF0R", "s43p", "9oqb", "ciZG", "1miu", "Z9L6", "HmT2", "-kRi"),
        )
        relatedRSOsHelper("QvrN0Z8ldkoOemOpqSA9azlQEGg", 0, emptyList())
        relatedRSOsHelper(
            "pDwUif48gNVEKIL_TsuMEtXYKyk",
            8,
            listOf("PPSY", "SvzT", "_5cc", "7U5p", "BUVr", "RRAP", "jMCD", "sHIR"),
        )
        relatedRSOsHelper(
            "Z9L6-YhsCyIdPrJuDzxTmqjK7RA",
            8,
            listOf("ciZG", "HmT2", "8Ooz", "1miu", "YWhd", "H0DT", "QF0R", "Kxk3"),
        )
        relatedRSOsHelper(
            "-kyhx31pYyqtW49P1bySc7pXnbI",
            8,
            listOf("Q8JH", "CZMZ", "FtGp", "VDxJ", "a--m", "1miu", "061b", "Rx-G"),
        )
        relatedRSOsHelper(
            "skMt0nszL5bm_6TKnX8pkYtJeYc",
            8,
            listOf("eQd5", "qWc7", "efE4", "7zCc", "5WQ1", "7Jsx", "lWVa", "8nVP"),
        )
        relatedRSOsHelper(
            "REAKQmiI3MgK1ZaIUgA6fWHVa30",
            8,
            listOf("nSms", "h836", "5sNA", "9oqb", "CZMZ", "osAD", "8nVP", "AhFn"),
        )
        relatedRSOsHelper(
            "1xPnHD0cb-_Wxq2mL1dnPAmQ0UA",
            8,
            listOf("9oqb", "SX3L", "90WJ", "sRzp", "W1x8", "XCf6", "3mFN", "S5k_"),
        )
        relatedRSOsHelper(
            "VDxJbfmKQgbxvw7IyABX3wud5Tg",
            8,
            listOf("a--m", "1miu", "MJPg", "L6Zl", "ATgT", "0Ibj", "b5f4", "eQd5"),
        )
        relatedRSOsHelper(
            "2OR8JRZMn_mqN3Gfi7bPj0AA97k",
            8,
            listOf("eQd5", "Q8JH", "ciZG", "BUVr", "VGdH", "HmT2", "lF3h", "gx81"),
        )
        relatedRSOsHelper(
            "E408COXofsxRMvYQnaGryf82Q5Y",
            8,
            listOf("eQd5", "qWc7", "efE4", "7zCc", "5WQ1", "7Jsx", "lWVa", "8nVP"),
        )
        relatedRSOsHelper(
            "8DKedINfeADfkFLDdendS0U22dU",
            8,
            listOf("nSms", "h836", "5sNA", "9oqb", "CZMZ", "osAD", "8nVP", "AhFn"),
        )
        relatedRSOsHelper("WkI_lm-om7kLYHDS5UKxFKEKwA8", 0, emptyList())
        relatedRSOsHelper(
            "zOyQ59-nfGOOW1db-BO0TI5E-AI",
            8,
            listOf("ciZG", "VGdH", "HmT2", "LfmV", "GqRf", "FP52", "HTKc", "GQef"),
        )
        relatedRSOsHelper(
            "ECpMsXd2o44YQJeZNwBL9Sbm4FY",
            8,
            listOf("ciZG", "1miu", "Z9L6", "HmT2", "QF0R", "ck_f", "8Ooz", "s43p"),
        )
        relatedRSOsHelper(
            "f87_Me-nb1CAps5zWoLhS1BDZuU",
            8,
            listOf("SvzT", "CnYg", "Q8JH", "iVc3", "Z0o5", "iUwL", "tqO2", "_5cc"),
        )
        relatedRSOsHelper("gtm_LE1sziMGkf5t72DC8uTmy7o", 0, emptyList())
        relatedRSOsHelper("zeG1dFQgvg05IlZk5nutgok9syI", 0, emptyList())
        relatedRSOsHelper("ZPmEZWbvsVwOHm72rI3wzTI0CuI", 0, emptyList())
        relatedRSOsHelper(
            "t0Plky6vzRKCckq5y4hT_8g8j4U",
            8,
            listOf("s43p", "2FO0", "SSws", "HmT2", "yA2n", "7U5p", "k2ZX", "5gES"),
        )
        relatedRSOsHelper(
            "Kxk3Dm7a2D1Jt7JFdrBM66qtJ14",
            8,
            listOf("s43p", "GqRf", "ciZG", "mx9h", "YWhd", "Z9L6", "HmT2", "Y0_A"),
        )
        relatedRSOsHelper(
            "1b-2JmM57W-6XAyCUXk9QGK6NaQ",
            8,
            listOf("eQd5", "XuEp", "EBpj", "AzBh", "zwzD", "NG7P", "yMXt", "GwoD"),
        )
        relatedRSOsHelper(
            "nSLaeZQdVjHT0A5YEljfNUNgo-Y",
            8,
            listOf("8nVP", "4XwG", "pW8m", "RhQ9", "uncP", "nZjK", "9rkE", "HUGz"),
        )
        relatedRSOsHelper(
            "G7SR0Ksn8ExsI-j2lWEc9m_JsJQ",
            8,
            listOf("h836", "7zCc", "Z0o5", "9oqb", "5WQ1", "lWVa", "oHiC", "ciZG"),
        )
        relatedRSOsHelper(
            "5Dkir_p8pApN0hcPbd_tqZ1xMOg",
            8,
            listOf("SvzT", "CnYg", "Q8JH", "iVc3", "Z0o5", "iUwL", "tqO2", "_5cc"),
        )
        relatedRSOsHelper(
            "adqnvii602M5W9N4YrB9a1bPOY8",
            8,
            listOf("yHll", "_5cc", "Lr36", "dvEi", "YPZ6", "dqIx", "MJPg", "TXYP"),
        )
        relatedRSOsHelper("_1e5f8fXw5pO7yrm_1guNimUibo", 0, emptyList())
        relatedRSOsHelper(
            "VPgk1IbTKAOzyTlHDA0nk1y9iv4",
            8,
            listOf("SvzT", "CnYg", "Q8JH", "iVc3", "Z0o5", "iUwL", "tqO2", "_5cc"),
        )
        relatedRSOsHelper(
            "XSIJ66W_bKsWGf05bXDzL526MHg",
            8,
            listOf("nSms", "h836", "9oqb", "8nVP", "AhFn", "urdk", "Bo1L", "FV6P"),
        )
        relatedRSOsHelper(
            "I47wiWWqg8Uj1qzMVmW4U0giH2w",
            8,
            listOf("SvzT", "CnYg", "Q8JH", "iVc3", "Z0o5", "iUwL", "tqO2", "_5cc"),
        )
        relatedRSOsHelper(
            "Qc1i2riTbh5Lg2kLR-KitLDFpCw",
            8,
            listOf("OGh2", "KFmr", "aGaa", "NW1z", "rdsh", "puxn", "es_D", "6yjJ"),
        )
        relatedRSOsHelper(
            "W1x8_mt1KWZMhHSQ5AisuBLcLY4",
            8,
            listOf("9oqb", "SX3L", "90WJ", "sRzp", "1xPn", "XCf6", "3mFN", "S5k_"),
        )
        relatedRSOsHelper("jAfQcMqe4xCVY7lrLS3iEaPqEtw", 0, emptyList())
        relatedRSOsHelper("mBLB6Gwr5SMpsHL_QfP6qM_Af0s", 0, emptyList())
        relatedRSOsHelper(
            "fYmt2yCz0BOl4zd_7thuZSvih7U",
            8,
            listOf("xkz9", "GqRf", "06MU", "eQd5", "Q8JH", "ciZG", "ybnD", "BUVr"),
        )
        relatedRSOsHelper("Vrq6nRMtzrNqiLQnOa6Tn0Qor70", 0, emptyList())
        relatedRSOsHelper("tn3lDLqhIDNEeVf41sWI0cYhSto", 0, emptyList())
        relatedRSOsHelper(
            "p6G0sCDUuTcFq161dv3wsgfplo0",
            8,
            listOf("9oqb", "HmT2", "QF0R", "Y0_A", "pDSa", "ck_f", "s43p", "j3pV"),
        )
        relatedRSOsHelper(
            "W95zIllqcBMflWvbpnBWttLd3lc",
            8,
            listOf("CZMZ", "061b", "SkTI", "x4_t", "zaMP", "TTKn", "jE_h", "hC77"),
        )
        relatedRSOsHelper(
            "MJPgAMW6aavHpoQ64xu5nVWUev8",
            8,
            listOf("9oqb", "1miu", "exMc", "VDxJ", "a--m", "7U5p", "mx9h", "j8ph"),
        )
        relatedRSOsHelper(
            "yf3SG9VFM2AEOGIPtnsb0x8TXnY",
            8,
            listOf("nSms", "h836", "9oqb", "8nVP", "AhFn", "urdk", "Bo1L", "FV6P"),
        )
        relatedRSOsHelper(
            "7DEOmnsyi_llJEq8EJHP4AOgT_Y",
            8,
            listOf("q-_r", "V4O3", "uUgY", "NAzx", "rVWl", "Ply2", "ZEVD", "6lVH"),
        )
        relatedRSOsHelper(
            "yxXiFPhKEQeBgBTKBfHrFuuMzDs",
            8,
            listOf("eQd5", "9oqb", "TllZ", "VDxJ", "ciZG", "a--m", "7U5p", "k2ZX"),
        )
        relatedRSOsHelper("ePpWEBsCP9vVSWvVFkdFm3dHRfw", 0, emptyList())
        relatedRSOsHelper(
            "vkF1bjzByeqZLKZPcpLN4pzS5T0",
            8,
            listOf("in6m", "rKUE", "iVc3", "5WQ1", "C75l", "_5cc", "hvLA", "1miu"),
        )
        relatedRSOsHelper(
            "j3pVDF45lKroVA_0VbOvNyayUTA",
            8,
            listOf("s43p", "9oqb", "exMc", "nfX6", "1miu", "j8ph", "QF0R", "Kxk3"),
        )
        relatedRSOsHelper(
            "gbCVGkF_9Z97XRXPaMrp2YFU8bk",
            8,
            listOf("nULV", "D25G", "OGh2", "KFmr", "Qc1i", "w5ld", "aGaa", "HH-e"),
        )
        relatedRSOsHelper(
            "Q45huzHmfE-a4KqGH2cBqMdYFmU",
            8,
            listOf("h836", "9oqb", "FV6P", "KmoR", "WGcd", "TTKn", "tM2r", "hC77"),
        )
        relatedRSOsHelper(
            "53W1z8tTA-WltES-MzkZRtHgA5U",
            8,
            listOf("s43p", "exMc", "HmT2", "RRAP", "GHDG", "9AmN", "1e5c", "PPSY"),
        )
        relatedRSOsHelper(
            "BW7E7Je-emoPd_NLcD_ZeuKsN3E",
            8,
            listOf("SvzT", "CnYg", "Q8JH", "iVc3", "Z0o5", "iUwL", "tqO2", "_5cc"),
        )
    }

    /** Test related RSOs shown in RSO view.  */
    @Test(timeout = 30000L)
    @Graded(points = 5, friendlyName = "Related RSOs UI (Integration)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test5_RelatedRSOsUI() {
        // Related RSO test must work first
        test4_testRelatedRSOs()

        // Get a shuffled list of RSO summaries
        val trimmedSummaries = getShuffledSummaries(12437, 32)

        // Use the random RSOs we picked above
        // Can't use the for-in loop here since we maintain two counters
        var count = 0
        var i = 0
        while (i < 32 && count < 4) {
            val summary = trimmedSummaries[i]

            // Skip RSOs without any relations
            val rso = testClient { callback -> Client.getRSO(summary.id, callback) }
            if (rso.relatedRSOs.isEmpty()) {
                i++
                continue
            }
            count++

            // Create the Intent
            val intent = Intent(ApplicationProvider.getApplicationContext(), RSOActivity::class.java).apply {
                putExtra("id", summary.id)
            }

            // Start the RSO activity
            startActivity<RSOActivity>(intent) {
                // Wait for it to load
                pause()
                // Check that the related RSOs are displayed
                for (related in rso.relatedRSOs) {
                    onView(withText(containsString(related.title))).check(matches(isDisplayed()))
                }
            }
            i++
        }
    }

    companion object {
        // Security manager used to count access to rsos.json
        private val jsonReadCountSecurityManager = JSONReadCountSecurityManager()

        // Run once before any test in this suite is started
        @JvmStatic
        @BeforeClass
        fun beforeClass() {
            // Check Server.kt for publicly visible members
            doCheckServerDesign()
            // Set up logging so that you can see log output during testing
            configureLogging()
            // Force loads to perform initialization before we start counting
            RSODATA.size
            SUMMARIES.size
            // Install our security manager that allows counting access to rsos.json
            System.setSecurityManager(jsonReadCountSecurityManager)
            // Start the API server
            startServer()
        }

        // Run once after all tests in this suite are completed
        @JvmStatic
        @AfterClass
        fun removeSecurityManager() {
            System.setSecurityManager(null)
        }
    }

    // Run before each test
    @Before
    fun beforeEachTest() {
        // Randomly set several RSOs to favorites before server reset to defend against client-side caching
        val trimmedSummaries = getShuffledSummaries(Random().nextInt(), 16)
        for (summary in trimmedSummaries) {
            try {
                @Suppress("UNUSED_VARIABLE")
                val ignored =
                    testClient { callback -> Client.setFavorite(summary.id, true, callback) }
            } catch (_: Exception) {}
        }
        // Reset the server before each test
        try {
            resetServer()
        } catch (_: Exception) {
        }
        // Check for extra reads from rsos.json
        jsonReadCountSecurityManager.checkCount()
    }

    // Run after each test
    @After
    fun afterEachTest() {
        // Check for extra reads from rsos.json
        jsonReadCountSecurityManager.checkCount()
    }
}

// md5: def5c1f647129069bc044eaddbb9fd4e // DO NOT REMOVE THIS LINE
